#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This file is part of PLCOpenEditor, a library implementing an IEC 61131-3 editor
#based on the plcopen standard. 
#
#Copyright (C): Edouard TISSERANT and Laurent BESSARD
#
#See COPYING file for copyrights details.
#
#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.
#
#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.
#
#You should have received a copy of the GNU General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from plcopen import plcopen
from plcopen.structures import *
from types import *

varTypeNames = {"localVars" : "VAR", "tempVars" : "VAR_TEMP", "inputVars" : "VAR_INPUT", 
                "outputVars" : "VAR_OUTPUT", "inOutVars" : "VAR_IN_OUT", "externalVars" : "VAR_EXTERNAL",
                "globalVars" : "VAR_GLOBAL", "accessVars" : "VAR_ACCESS"}

pouTypeNames = {"function" : "FUNCTION", "functionBlock" : "FUNCTION_BLOCK", "program" : "PROGRAM"}

def ReIndentText(text, nb_spaces):
    compute = ""
    lines = text.splitlines()
    if len(lines) > 0:
        spaces = 0
        while lines[0][spaces] == " ":
            spaces += 1
        indent = ""
        for i in xrange(spaces, nb_spaces):
            indent += " "
        for line in lines:
            if line != "":
                compute += "%s%s\n"%(indent, line)
            else:
                compute += "\n"
    return compute

"""
Module implementing methods for generating PLC programs in ST or IL
"""

class PouProgram:
    
    def __init__(self, name, type):
        self.Name = name
        self.Type = type
        self.ReturnType = None
        self.Interface = []
        self.InitialSteps = []
        self.SFCNetworks = {"Steps":{}, "Transitions":{}, "Actions":{}}
        self.Program = ""
    
    def IsAlreadyDefined(self, name):
        for list_type, retain, constant, vars in self.Interface:
            for var_type, var_name in vars:
                if name == var_name:
                    return True
        return False
    
    def GenerateInterface(self, interface):
        if self.Type == "FUNCTION":
            self.ReturnType = interface.getReturnType().getValue()
        for varlist in interface.getContent():
            variables = []
            for var in varlist["value"].getVariable():
                type = var.getType().getValue()
                variables.append((type, var.getName()))
            self.Interface.append((varTypeNames[varlist["name"]], varlist["value"].getRetain(), 
                            varlist["value"].getConstant(), variables))
    
    def GenerateProgram(self, pou):
        body = pou.getBody()
        body_content = body.getContent()
        body_type = body_content["name"]
        if body_type in ["IL","ST"]:
            self.Program = ReIndentText(body_content["value"].getText(), 2)
        elif body_type == "FBD":
            for instance in body.getContentInstances():
                if isinstance(instance, plcopen.outVariable):
                    var = instance.getExpression()
                    connections = instance.connectionPointIn.getConnections()
                    if connections and len(connections) == 1:
                        expression = self.ComputeFBDExpression(body, connections[0])
                        self.Program += "  %s := %s;\n"%(var, expression)
        elif body_type == "LD":
            for instance in body.getContentInstances():
                if isinstance(instance, plcopen.coil):
                    paths = self.GenerateLDPaths(instance, body)
                    variable = self.ExtractModifier(instance, instance.getVariable())
                    expression = self.ComputeLDExpression(paths, True)
                    self.Program += "  %s := %s;\n"%(variable, expression)
        elif body_type == "SFC":
            for instance in body.getContentInstances():
                if isinstance(instance, plcopen.step):
                    self.GenerateSFCStep(instance, pou)
                elif isinstance(instance, plcopen.actionBlock):
                    self.GenerateSFCStepActions(instance, pou)
                elif isinstance(instance, plcopen.transition):
                    self.GenerateSFCTransition(instance, pou)
                elif isinstance(instance, plcopen.jumpStep):
                    self.GenerateSFCJump(instance, pou)
            for initialstep in self.InitialSteps:
                self.ComputeSFCStep(initialstep)
    
    def ComputeFBDExpression(self, body, link):
        localid = link.getRefLocalId()
        instance = body.getContentInstance(localid)
        if isinstance(instance, plcopen.inVariable):
            return instance.getExpression()
        elif isinstance(instance, plcopen.block):
            name = instance.getInstanceName()
            type = instance.getTypeName()
            block_infos = GetBlockType(type)
            if block_infos["type"] == "function":
                vars = []
                for variable in instance.inputVariables.getVariable():
                    connections = variable.connectionPointIn.getConnections()
                    if connections and len(connections) == 1:
                        value = self.ComputeFBDExpression(body, connections[0])
                        vars.append(self.ExtractModifier(variable, value))
                variable = instance.outputVariables.getVariable()[0]
                return self.ExtractModifier(variable, "%s(%s)"%(type, ", ".join(vars)))
            elif block_infos["type"] == "functionBlock":
                if self.Interface[-1][0] != "VAR" or self.Interface[-1][1] or self.Interface[-1][2]:
                    self.Interface.append(("VAR", False, False, []))
                if not self.IsAlreadyDefined(name):
                    self.Interface[-1][3].append((type, name))
                    vars = []
                    for variable in instance.inputVariables.getVariable():
                        connections = variable.connectionPointIn.getConnections()
                        if connections and len(connections) == 1:
                            parameter = variable.getFormalParameter()
                            value = self.ComputeFBDExpression(body, connections[0])
                            vars.append(self.ExtractModifier(variable, "%s := %s"%(parameter, value)))
                    self.Program += "  %s(%s);\n"%(name, ", ".join(vars))
                connectionPoint = link.getPosition()[-1]
                for variable in instance.outputVariables.getVariable():
                    blockPointx, blockPointy = variable.connectionPointOut.getRelPosition()
                    if instance.getX() + blockPointx == connectionPoint.getX() and instance.getY() + blockPointy == connectionPoint.getY():
                        return self.ExtractModifier(variable, "%s.%s"%(name, variable.getFormalParameter()))
                raise ValueError, "No output variable found"

    def GenerateLDPaths(self, instance, body):
        paths = []
        variable = self.ExtractModifier(instance, instance.getVariable())
        connections = instance.connectionPointIn.getConnections()
        for connection in connections:
            localId = connection.getRefLocalId()
            next = body.getContentInstance(localId)
            if isinstance(next, plcopen.leftPowerRail):
                paths.append(None)
            else:
                paths.append(self.GenerateLDPaths(next, body))
        if isinstance(instance, plcopen.coil):
            if len(paths) > 1:
                return tuple(paths)
            else:
                return paths
        else:
            if len(paths) > 1:
                return [variable, tuple(paths)]
            elif type(paths[0]) == ListType:
                return [variable] + paths[0]
            elif paths[0]:
                return [variable, paths[0]]
            else:
                return variable

    def ExtractDivergenceInput(self, divergence, pou):
        connectionPointIn = divergence.getConnectionPointIn()
        if connectionPointIn:
            connections = connectionPointIn.getConnections()
            if len(connections) == 1:
                instanceLocalId = connections[0].getRefLocalId()
                return pou.body.getContentInstance(instanceLocalId)
        return None

    def ExtractConvergenceInputs(self, convergence, pou):
        instances = []
        for connectionPointIn in convergence.getConnectionPointIn():
            connections = connectionPointIn.getConnections()
            if len(connections) == 1:
                instanceLocalId = connections[0].getRefLocalId()
                instances.append(pou.body.getContentInstance(instanceLocalId))
        return instances

    def GenerateSFCStep(self, step, pou):
        step_name = step.getName()
        if step_name not in self.SFCNetworks["Steps"].keys():
            if step.getInitialStep():
                self.InitialSteps.append(step_name)
            step_infos = {"initial" : step.getInitialStep(), "transitions" : [], "actions" : []}
            if step.connectionPointIn:
                instances = []
                connections = step.connectionPointIn.getConnections()
                if len(connections) == 1:
                    instanceLocalId = connections[0].getRefLocalId()
                    instance = pou.body.getContentInstance(instanceLocalId)
                    if isinstance(instance, plcopen.transition):
                        instances.append(instance)
                    elif isinstance(instance, plcopen.selectionConvergence):
                        instances.extend(self.ExtractConvergenceInputs(instance, pou))
                    elif isinstance(instance, plcopen.simultaneousDivergence):
                        transition = self.ExtractDivergenceInput(instance, pou)
                        if transition:
                            if isinstance(transition, plcopen.transition):
                                instances.append(transition)
                            elif isinstance(transition, plcopen.selectionConvergence):
                                instances.extend(self.ExtractConvergenceInputs(transition, pou))
                for instance in instances:
                    self.GenerateSFCTransition(instance, pou)
                    if instance in self.SFCNetworks["Transitions"].keys():
                        self.SFCNetworks["Transitions"][instance]["to"].append(step_name)
            self.SFCNetworks["Steps"][step_name] = step_infos
    
    def GenerateSFCJump(self, jump, pou):
        jump_target = jump.getTargetName()
        if jump.connectionPointIn:
            instances = []
            connections = jump.connectionPointIn.getConnections()
            if len(connections) == 1:
                instanceLocalId = connections[0].getRefLocalId()
                instance = pou.body.getContentInstance(instanceLocalId)
                if isinstance(instance, plcopen.transition):
                    instances.append(instance)
                elif isinstance(instance, plcopen.selectionConvergence):
                    instances.extend(self.ExtractConvergenceInputs(instance, pou))
                elif isinstance(instance, plcopen.simultaneousDivergence):
                    transition = self.ExtractDivergenceInput(instance, pou)
                    if transition:
                        if isinstance(transition, plcopen.transition):
                            instances.append(transition)
                        elif isinstance(transition, plcopen.selectionConvergence):
                            instances.extend(self.ExtractConvergenceInputs(transition, pou))
            for instance in instances:
                self.GenerateSFCTransition(instance, pou)
                if instance in self.SFCNetworks["Transitions"].keys():
                    self.SFCNetworks["Transitions"][instance]["to"].append(jump_target)
    
    def GenerateSFCStepActions(self, actionBlock, pou):
        connections = actionBlock.connectionPointIn.getConnections()
        if len(connections) == 1:
            stepLocalId = connections[0].getRefLocalId()
            step = pou.body.getContentInstance(stepLocalId)
            self.GenerateSFCStep(step, pou)
            step_name = step.getName()
            if step_name in self.SFCNetworks["Steps"].keys():
                actions = actionBlock.getActions()
                for action in actions:
                    action_infos = {"qualifier" : action["qualifier"], "content" : action["value"]}
                    if "duration" in action:
                        action_infos["duration"] = action["duration"]
                    if "indicator" in action:
                        action_infos["indicator"] = action["indicator"]
                    if action["type"] == "reference":
                        self.GenerateSFCAction(action["value"], pou)
                    self.SFCNetworks["Steps"][step_name]["actions"].append(action_infos)
    
    def GenerateSFCAction(self, action_name, pou):
        if action_name not in self.SFCNetworks["Actions"].keys():
            actionContent = pou.getAction(action_name)
            if actionContent:
                actionType = actionContent.getBodyType()
                actionBody = actionContent.getBody()
                if actionType in ["ST", "IL"]:
                    self.SFCNetworks["Actions"][action_name] = ReIndentText(actionContent.getText(), 4)
                elif actionType == "FBD":
                    for instance in actionBody.getContentInstances():
                        if isinstance(instance, plcopen.outVariable):
                            var = instance.getExpression()
                            connections = instance.connectionPointIn.getConnections()
                            if connections and len(connections) == 1:
                                expression = self.ComputeFBDExpression(actionBody, connections[0])
                                self.SFCNetworks["Actions"][action_name] = self.Program + "  %s := %s;\n"%(var, expression)
                                self.Program = ""
                elif actionType == "LD":
                    for instance in actionbody.getContentInstances():
                        if isinstance(instance, plcopen.coil):
                            paths = self.GenerateLDPaths(instance, actionBody)
                            variable = self.ExtractModifier(instance, instance.getVariable())
                            expression = self.ComputeLDExpression(paths, True)
                            self.SFCNetworks["Actions"][action_name] = self.Program + "  %s := %s;\n"%(variable, expression)
                            self.Program = ""
    
    def GenerateSFCTransition(self, transition, pou):
        if transition not in self.SFCNetworks["Transitions"].keys():
            steps = []
            connections = transition.connectionPointIn.getConnections()
            if len(connections) == 1:
                instanceLocalId = connections[0].getRefLocalId()
                instance = pou.body.getContentInstance(instanceLocalId)
                if isinstance(instance, plcopen.step):
                    steps.append(instance)
                elif isinstance(instance, plcopen.selectionDivergence):
                    step = self.ExtractDivergenceInput(instance, pou)
                    if step:
                        if isinstance(step, plcopen.step):
                            steps.append(step)
                        elif isinstance(step, plcopen.simultaneousConvergence):
                            steps.extend(self.ExtractConvergenceInputs(step, pou))
                elif isinstance(instance, plcopen.simultaneousConvergence):
                    steps.extend(self.ExtractConvergenceInputs(instance, pou))
            transition_infos = {"from": [], "to" : []}
            transitionValues = transition.getConditionContent()
            if transitionValues["type"] == "inline":
                transition_infos["content"] = "\n    := %s;\n"%transitionValues["value"]
            else:
                transitionContent = pou.getTransition(transitionValues["value"])
                transitionType = transitionContent.getBodyType()
                transitionBody = transitionContent.getBody()
                if transitionType == "IL":
                    transition_infos["content"] = ":\n%s"%ReIndentText(transitionBody.getText(), 4)
                elif transitionType == "ST":
                    transition_infos["content"] = "\n%s"%ReIndentText(transitionBody.getText(), 4)
                elif conditionType == "FBD":
                    for instance in transitionBody.getContentInstances():
                        if isinstance(instance, plcopen.outVariable):
                            connections = instance.connectionPointIn.getConnections()
                            if connections and len(connections) == 1:
                                expression = self.ComputeFBDExpression(actionBody, connections[0])
                                transition_infos["content"] = "\n    := %s;\n"%(var, expression)
                elif actionType == "LD":
                    for instance in transitionbody.getContentInstances():
                        if isinstance(instance, plcopen.coil):
                            paths = self.GenerateLDPaths(instance, conditionBody)
                            expression = self.ComputeLDExpression(paths, True)
                            transition_infos["content"] = "\n    := %s;\n"%expression
            for step in steps:
                self.GenerateSFCStep(step, pou)
                step_name = step.getName()
                if step_name in self.SFCNetworks["Steps"].keys():
                    transition_infos["from"].append(step_name)
                    self.SFCNetworks["Steps"][step_name]["transitions"].append(transition)
            self.SFCNetworks["Transitions"][transition] = transition_infos

    def ComputeSFCStep(self, step_name):
        if step_name in self.SFCNetworks["Steps"].keys():
            step_infos = self.SFCNetworks["Steps"].pop(step_name)
            if step_infos["initial"]:
                self.Program += "  INITIAL_STEP %s:\n"%step_name
            else:
                self.Program += "  STEP %s:\n"%step_name
            actions = []
            for action_infos in step_infos["actions"]:
                actions.append(action_infos["content"])
                self.Program += "    %(content)s(%(qualifier)s"%action_infos
                if "duration" in action_infos:
                    self.Program += ", %(duration)s"%action_infos
                if "indicator" in action_infos:
                    self.Program += ", %(indicator)s"%action_infos
                self.Program += ");\n"
            self.Program += "  END_STEP\n\n"
            for action in actions:
                self.ComputeSFCAction(action)
            for transition in step_infos["transitions"]:
                self.ComputeSFCTransition(transition)
                
    def ComputeSFCAction(self, action_name):
        if action_name in self.SFCNetworks["Actions"].keys():
            action_content = self.SFCNetworks["Actions"].pop(action_name)
            self.Program += "  ACTION %s:\n%s  END_ACTION\n\n"%(action_name, action_content)
    
    def ComputeSFCTransition(self, transition):
        if transition in self.SFCNetworks["Transitions"].keys():
            transition_infos = self.SFCNetworks["Transitions"].pop(transition)
            self.Program += "  TRANSITION FROM "
            if len(transition_infos["from"]) > 1:
                self.Program += "(%s)"%", ".join(transition_infos["from"])
            else:
                self.Program += "%s"%transition_infos["from"][0]
            self.Program += " TO "
            if len(transition_infos["to"]) > 1:
                self.Program += "(%s)"%", ".join(transition_infos["to"])
            else:
                self.Program += "%s"%transition_infos["to"][0]
            self.Program += transition_infos["content"]
            self.Program += "  END_TRANSITION\n\n"
            for step_name in transition_infos["to"]:
                self.ComputeSFCStep(step_name)
        
    def ComputeLDExpression(self, paths, first = False):
        if type(paths) == TupleType:
            if None in paths:
                return "TRUE"
            else:
                var = [self.ComputeLDExpression(path) for path in paths]
                if first:
                    return " OR ".join(var)
                else:
                    return "(%s)"%" OR ".join(var)
        elif type(paths) == ListType:
            var = [self.ComputeLDExpression(path) for path in paths]
            return " AND ".join(var)
        else:
            return paths

    def ExtractModifier(self, variable, text):
        if variable.getNegated():
            return "NOT(%s)"%text
        else:
            edge = variable.getEdge()
            if edge:
                if edge.getValue() == "rising":
                    return self.AddTrigger("R_TRIG", text)
                elif edge.getValue() == "falling":
                    return self.AddTrigger("F_TRIG", text)
        return text
    
    def AddTrigger(self, edge, text):
        if self.Interface[-1][0] != "VAR" or self.Interface[-1][1] or self.Interface[-1][2]:
            self.Interface.append(("VAR", False, False, []))
        i = 1
        name = "%s%d"%(edge, i)
        while self.IsAlreadyDefined(name):
            i += 1
            name = "%s%d"%(edge, i)
        self.Interface[-1][3].append((edge, name))
        self.Program += "  %s(CLK := %s);\n"%(name, text)
        return "%s.Q"%name
    
    def GenerateSTProgram(self):
        program = ""
        if self.ReturnType:
            program += "%s %s : %s\n"%(self.Type, self.Name, self.ReturnType)
        else:
            program += "%s %s\n"%(self.Type, self.Name)
        for list_type, retain, constant, variables in self.Interface:
            program += "  %s"%list_type
            if retain:
                program += " RETAIN"
            if constant:
                program += " CONSTANT"
            program += "\n"
            for var_type, var_name in variables:
                program += "    %s : %s;\n"%(var_name, var_type)
            program += "  END_VAR\n"
        program += "\n"
        program += self.Program
        program += "END_%s\n\n"%self.Type
        return program
    
def GenerateCurrentProgram(project):
    program = ""
    for pou in project.getPous():
        pou_type = pou.getPouType().getValue()
        if pou_type in pouTypeNames:
            pou_program = PouProgram(pou.getName(), pouTypeNames[pou_type])
        else:
            raise ValueError, "Undefined pou type"
        pou_program.GenerateInterface(pou.getInterface())
        pou_program.GenerateProgram(pou)
        program += pou_program.GenerateSTProgram()
    return program

